#!/usr/bin/perl

########################################################################
# metafsrc2raw.pl 1.42
#   convert data from different sources to standard format
#
# copyright (c) metaf2xml 2009-2010
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
########################################################################

########################################################################
# some things strictly perl
########################################################################
use strict;
use warnings;

=head1 NAME

metafsrc2raw.pl

=head1 SYNOPSIS

  metafsrc2raw.pl -F<format> [OPTION]... [FILE]...

=head1 DESCRIPTION

This script converts METAR, TAF, or SYNOP information provided in files on
various public Internet servers to messages delimited by a newline and
in the format specified by the I<WMO Manual No. 306>, without modifications
due to distribution (which e.g. lists the initial part of messages only once
for several messages, or uses the "=" (equal sign) as a delimiter).
It removes the leading keywords C<METAR>, C<SPECI>, C<TAF>, and C<SYNOP> and
prepends parts of the WMO header according to the option -H.

If no input files are specified, then the standard input is read.

=head1 DEPENDENCIES

The Perl module C<Getopt::Std> is required.

=head1 ABBREVIATIONS

=over

=item noaa

National Oceanic and Atmospheric Administration

=item nws

the National Weather Service department of the NOAA

=item iws

the Internet Weather Source of the NWS

=item adds

the Aviation Digital Data Service of the NWS

=item addsds

the experimental data server of the Aviation Digital Data Service of the NWS

=item cod

College of DuPage

=item fsu

Florida State University

=back

=head1 OPTIONS

=over

=item -F I<format>

origin of the input data: metar_cycle_iws, staf_cycle_iws, taf_cycle_iws, metaf_nws, metaf_cod, metaf_fsu, synop_nws, synop_cod, synop_fsu

=item -D

suppress duplicate messages

=item -N

suppress messages that only contain C<NIL>

=item -H I<0..4>

which parts of the WMO header (C<TTAA[II] CCCC DDHHMM> [modifiers]) to prepend
to each message:

=over

=item -H0 (default)

(empty header)

=item -H1

C<TT> (report type:

=over

=item SA

METAR

=item SP

SPECI

=item FT

TAF with forecast period >= 12 hours

=item FC

TAF with forecast period < 12 hours

=item SM

SYNOP at main hours (00:00, 06:00, 12:00, 18:00 UTC)

=item SI

SYNOP at intermediate (main + 3) hours

=item SN

SYNOP at non-standard (other than main and intermediate) hours

=back

)

=item -H2

C<TTAA[II]> (like C<-H1> plus region for the stations and optional report number)

=item -H3

C<TTAA[II] CCCC> (like C<-H2> plus report dissimination location)

=item -H4

C<TTAA[II] CCCC DDHHMM> (like C<-H3> plus report dissimination time)

=back

=back

=head1 EXAMPLES

Parse SYNOP messages from NOAA/NWS, suppressing NIL messages and duplicates:

  metafsrc2raw.pl -DNF synop_nws <<"EOF"
####018000228####
SMPY01 SGAS 060600
AAXX 06064
86134 32970 02402 10250 20235 40105 56005=
86297 NIL=
86134 32970 02402 10250 20235 40105 56005=
86218 32970 01810 10240 20209 40103 58011=
####018000158####
SMFG01 SOCA 060600
AAXX 06064
81405 01259 60505 10241 20229 30105 40114 60071 76086 86200
  333 32024 55048 60067 58002 70098 82705 86813
  555 69905=####018000118####
SMCS01 MROC 060600
AAXX 06064
78762 32970 00915 10220 20160 39094 40101 57005
333 10280 20193 30/// 59006=
EOF

results in:

  AAXX 06064 86134 32970 02402 10250 20235 40105 56005
  AAXX 06064 86218 32970 01810 10240 20209 40103 58011
  AAXX 06064 81405 01259 60505 10241 20229 30105 40114 60071 76086 86200 333 32024 55048 60067 58002 70098 82705 86813 555 69905
  AAXX 06064 78762 32970 00915 10220 20160 39094 40101 57005 333 10280 20193 30/// 59006

Parse METAR/SPECI messages from the College of DuPage with the WMO header
C<TTAAII> prepended:

  metafsrc2raw.pl -F metaf_cod -H2 <<"EOF"
355 
SAUS70 KWBC 161200
METAR
KDZJ 161146Z AUTO 00000KT 10SM -DZ SCT002 BKN075 OVC100 09/09
     A3013 RMK AO2=
KLDJ 161150Z AUTO 00000KT 10SM CLR 03/M01 A3027 RMK AO2=
KMLJ 161145Z AUTO 00000KT 4SM RA BKN005 BKN012 OVC033 12/12 A3017
     RMK AO2=
351 
SPCN51 CWAO 161254
SPECI CWCL 161254Z AUTO 19008KT 9SM FEW008 FEW015 BKN027 BKN043
M04/M06 A2947=
EOF

results in:

  SAUS70 KDZJ 161146Z AUTO 00000KT 10SM -DZ SCT002 BKN075 OVC100 09/09 A3013 RMK AO2
  SAUS70 KLDJ 161150Z AUTO 00000KT 10SM CLR 03/M01 A3027 RMK AO2
  SAUS70 KMLJ 161145Z AUTO 00000KT 4SM RA BKN005 BKN012 OVC033 12/12 A3017 RMK AO2
  SPCN51 CWCL 161254Z AUTO 19008KT 9SM FEW008 FEW015 BKN027 BKN043 M04/M06 A2947

=head1 SEE ALSO

=begin html

<a href=metaf2xml.pl.html>metaf2xml(1)</a>,
<a href=metaf.pl.html>metaf(1)</a>, <!--

=end html

B<metaf2xml>(1),
B<metaf>(1),

=for html -->

L<http://metaf2xml.sourceforge.net/>

=head1 COPYRIGHT and LICENSE

copyright (c) 2009-2010 metaf2xml @ L<http://metaf2xml.sourceforge.net/>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA

=cut

########################################################################
# define some regular expressions
########################################################################
my $re_id_synop = '(?:S[IMN][A-Z]{2}\d{2})';
my $re_id_metaf = '(?:(?:S[AP]|F[CT])[A-Z]{2}\d{2})';
my $re_day  = '(?:0[1-9]|[12]\d|3[01])';
my $re_hour = '(?:[01]\d|2[0-3])';
my $re_min  = '[0-5]\d';
my $re_header_synop = '(?:AAXX {0,2}'. $re_day . $re_hour .'[0134/]|BBXX|OOXX)';
my $re_ICAO = '(?:[A-Z][A-Z\d]{3})';
my $re_header_metaf = "(?:(?:COR |AMD )*$re_ICAO )";

#
# output a message with optional check for duplicates
#
{
    my @msgs = ();

    sub print_msg {
        my ($msg, $no_dups) = @_;
        $msg =~ s/  +/ /g;
        $msg =~ s/[ =]*$//;
        if ($no_dups) {
            return if grep { $_ eq $msg } @msgs;
            push @msgs, $msg;
        }
        print "$msg\n";
        return;
    }
}

# special cases:
#   TAFPA TAF PHLI 292334Z ...
#   TAFPQ TAF PGUM 292347Z ...
#   ONTAF VTSP 300330Z ...
#   TFFR 300600Z METAR VRB01KT ...
#   TX OPMET V1
sub print_metaf_msg {
    my ($wmo_header, $msg, $opts) = @_;

    $msg =~ s/  +/ /g;
    $msg =~ s/^ +//;
    $msg =~ s/ +$//;
    $msg =~ s/^(?:PART \d+ OF \d+ PARTS |ONTAF |TAF(?:P[AQ])? |METAR |SPECI )+//;
    $msg =~ s/^(?:AMD |TAF )+/AMD /;
    $msg =~ s/^(?:COR |TAF )+/COR /;

    print_msg "$wmo_header$msg", exists $opts->{D}
        unless exists $opts->{N}
           && $msg =~ /^$re_header_metaf(?:$re_day$re_hour${re_min}Z (?:$re_day(?:$re_hour|24){2} )?)?NIL[= ]*$/;
    return;
}

# WMO header: TTAA[II] CCCC DDHHMM [modifiers]
#   TT:     report type (e.g. SM: main synoptic hour reports, SA: METAR)
#   AA:     region for the stations in the report
#   II:     report number (optional)
#   CCCC:   the report dissimination location
#   DDHHMM: the report dissimination time
#
#   $opt_H: 0 -> (empty)
#           1 -> TT
#           2 -> TTAA[II]
#           3 -> TTAA[II] CCCC
#           4 -> TTAA[II] CCCC DDHHMM
sub format_wmo_header {
    my ($hdr, $opt_H) = @_;

    return '' unless $opt_H > 0;
    return ($hdr =~ /((((..)[^ ]+) [^ ]+) \d+)/)[4 - $opt_H] . ' ';
}

#
# parse METAR/TAF file from NOAA/NWS, College of DuPage, or Florida State Univ.
#   difference: the pattern for the SOM (apart from white space)
#
sub process_metaf_nws_cod_fsu {
    my ($opts, $re_SOM) = @_;
    my ($state, $wmo_header, $msg);

    $state      = '';
    $wmo_header = '';
    $msg        = '';
    while (<>) {
        s/[\n\r\cA\cC\cZ]+//g;

        # check for trailing encoded(!) '=' or ' '
        s/(?<=[^ ])=3D(?:=20)?=? *$/=/;
        s/(?<=[^ ])=20 *$//;

        next
            if /^TX OPMET V1$/;

        # (STARTED: MSGEND)?, -> SOM
        if (/$re_SOM/) {
            # final '=' missing in previous block or missing new line?
            if ($state eq 'STARTED' && ($msg ne '' || $1)) {
                $msg .= $1 if $1;
                print_metaf_msg $wmo_header, $msg, $opts
                    if $msg =~ /^$re_header_metaf *[^ ]{3,}/o;
                $msg = '';
            }
            $state = 'SOM';

        # SOM -> Sx
        # SOM: KAWN uses SA(GL|EU|EW|UK|..), FT(MX), ..
        } elsif (   $state eq 'SOM'
                 && (   /^$re_id_metaf [A-Z\d]{4} \d{6}/o
                     || /^(?:S[AP]|F[CT])[A-Z]{2} KAWN \d{6}/))
        {
            $state      = 'Sx';
            $wmo_header = format_wmo_header $_, $opts->{H};

        # Sx: K*, PANC, PGUM, PHFO, TJSJ use (MTR|TAF)[A-Z\d]{3}
        # Sx: UHHH, WMKK use 'METAR DDHHMMZ'
        } elsif (   $state eq 'Sx'
                 && (/^(?:MTR|TAF)[A-Z\d]{3}$/ || /^METAR \d{6}Z$/))
        {
            # skip this line

        # Sx, STARTED: (MSGSTART +) MSGEND -> Sx
        } elsif (($state eq 'STARTED' || $state eq 'Sx') && /= *$/) {
            if ($state eq 'STARTED' && /^(?:METAR|SPECI|TAF) /) {
                # new message starts, previous didn't end with =
                $msg =~ s/^(?:METAR +|SPECI +|TAF +)+//;
                print_metaf_msg $wmo_header, $msg, $opts
                    if $msg =~ /^$re_header_metaf *[^ ]{3,}/o;
                $msg = '';
            }
            /^(.*)= *$/;
            $msg = "$msg$1";
            $msg =~ s/^(?:METAR +|SPECI +|TAF +)+//;
            print_metaf_msg $wmo_header, $msg, $opts
                if $msg =~ /^$re_header_metaf *[^ ]{3,}/o;
            $msg = '';
            $state = 'Sx';

        # Sx, STARTED: MSGSTART -> STARTED
        } elsif (   (   ($state eq 'STARTED' || $state eq 'Sx')
                     && /^(?:METAR|SPECI|TAF)(?: +([^ ].*))?/)
                 || ($state eq 'Sx' && /^ *([^ ].*)/)) # METAR|SPECI|TAF missing
        {
            $msg = "$1 " if $1;
            $state = 'STARTED' if $msg;

        # STARTED: MSGCONTD
        } elsif ($state eq 'STARTED' && /^ *([^ ].*)/) {
            $msg .= "$1 ";
        }
    }
    # end of message could be missing in last block
    print_metaf_msg $wmo_header, $msg, $opts
        if $state eq 'STARTED' && $msg;
    return;
}

#
# parse SYNOP file from NOAA/NWS, College of DuPage or Florida State University
#   difference: the pattern for the SOM (apart from white space)
#
sub process_synop_nws_cod_fsu {
    my ($opts, $re_SOM) = @_;
    my ($state, $wmo_header, $msg, $header, $report_id, $dissim, $date);

    $state      = '';
    $wmo_header = '';
    $msg        = '';
    while (<>) {
        s/[\n\r\cA\cC\cZ]+//g;

        # check for trailing encoded(!) '=' or ' '
        s/(?<=[^ ])=3D(?:=20)?=? *$/=/;
        s/(?<=[^ ])=20 *$//;

        next
            if /^NNNN/ || (exists $opts->{N} && /NIL=$/);

        # (STARTED: MSGEND)? -> SOM
        if (/$re_SOM/) {
            # final '=' missing in previous block or missing new line?
            if ($state eq 'STARTED' && ($msg ne '' || $1)) {
                print_msg "$wmo_header$header$msg$1", exists $opts->{D};
                $msg = '';
            }
            $state = 'SOM';

        # ? -> Sx
        } elsif (/^($re_id_synop) ([A-Z\d]{4}) (\d{4})/o) {
            $state      = 'Sx';
            $report_id  = $1;
            $dissim     = $2;
            $date       = $3;
            $wmo_header = format_wmo_header $_, $opts->{H};

        # Sx, STARTED -> STARTED
        } elsif (   ($state eq 'Sx' || $state eq 'STARTED')
                 && /$re_header_synop(?: +[^ ].*| *)$/o)
        {
            print_msg "$wmo_header$header$msg", exists $opts->{D}
                if $msg;

            s/(?<=^AAXX) */ /;
            /($re_header_synop)(?: +([^ ].*)| *)$/o;
            $state = 'STARTED';
            if (defined $2) {
                $msg = "$1 $2 ";
                $header = '';
            } else {
                $msg = '';
                $header = "$1 ";
            }

        # Sx -> STARTED
        } elsif ($state eq 'Sx') {
            if ($dissim eq 'KGYX' && /^SSMMWN/) {
                # KGYX omits 'AAXX YYGGiw' (and has no '=')
                $state = 'STARTED';
                $msg = '';
                $header = "AAXX ${date}4 ";
            } elsif ($dissim eq 'PTYA' && /^SSYM2 (.{5})/) {
                # PTYA sometimes uses 'SSYM2' instead of 'AAXX' (and has no '=')
                $state = 'STARTED';
                $msg = '';
                $header = "AAXX ${date}4 ";
            }

        # STARTED: MSGSTART SxPF80 KWBC
        # SxPF80 KWBC does not use delimiters, check for IIiii + (NIL | iRixhVV)
        } elsif (   $state eq 'STARTED'
                 && $dissim eq 'KWBC'
                 && substr($report_id, 2) eq 'PF80'
                 && m@^919(?:2[59]|4[358]|5[48]) (?:NIL|[0-4/][1-7/][\d/]{3} )@)
        {
            print_msg "$wmo_header$header$msg", exists $opts->{D}
                if $msg;
            $msg = "$_ ";

        # STARTED: MSGEND
        # - MDSD uses empty lines as delimiter
        # - PTKK sometimes has no delimiters, last group is (555 .* 9YYGG | NIL)
        } elsif (   $state eq 'STARTED'
                 && (   /^(.*)= *$/                         # end of message
                     || ($dissim eq 'MDSD' && /^ *()$/)     # MDSD && empty line
                     || (   $dissim eq 'PTKK'               # PTKK && ...
                         && (   /^ *(.* NIL)\s*$/           #  ... NIL
                             || (   "$msg $_" =~ / 555 /    #  ... 555 .* 9YYGG
                                 && /^ *(.* 9$date)\s*$/)))))
        {
            $msg = "$header$msg$1";
            print_msg "$wmo_header$msg", exists $opts->{D}
                if $msg =~ /^$re_header_synop *[^ ]{3,}/o;
            $msg = '';

        # STARTED: MSGCONTD
        } elsif ($state eq 'STARTED' && /^ *(.+)/) {
            $msg .= $1;
            # MXBA uses fixed width format
            $msg .= ' ' unless $dissim eq 'MXBA';
        }
    }
    # end of message could be missing in last block
    print_msg "$wmo_header$header$msg", exists $opts->{D}
        if $state eq 'STARTED' && $msg;
    return;
}

#
# parse METAR/STAF/TAF cycle file from NOAA/IWS
#
# special cases:
#   TAF GCLP 292300Z 300606 01016KT CAVOK= TAF GCXO 292300Z ...
#   TAF HECA 300400Z 300606 VRB03KT ... CAVOK= HELX 300400Z ...
#   DRZR 300600Z 00000KT CAVOK 24/23 Q1013 METAR DNAA 300600Z ...
sub process_metaf_cycle_iws {
    my ($opts, $type) = @_;
    my ($wmo_header, $wmo_header_fix, $msg);

    $wmo_header_fix =   { metar => 'SA', staf => 'FC', taf => 'FT' }->{$type}
                      . 'XX NOAA ';
    $wmo_header = format_wmo_header "${wmo_header_fix}010000", $opts->{H};
    $msg  = '';
    while (<>) {
        s/[\n\r\cC\cZ]+//g;

        # check for trailing encoded(!) '=' or ' '
        s/(?<=[^ ])=3D(?:=20)?=? *$/=/;
        s/(?<=[^ ])=20 *$//;

        if (/^ *$/) {                                     # end of message?
            print_metaf_msg $wmo_header, $msg, $opts;
            $msg = '';
        } elsif (   /(.*)= ?TAF (.*)/                     # TAF after '='?
                 || /(.*)= ?($re_ICAO .*)/                # ICAO code after '='?
                 || /(.*) METAR ($re_ICAO .*)/)           # METAR + ICAO code?
        {
            print_metaf_msg $wmo_header, "$msg $1", $opts;
            $msg = $2;
        } elsif (/^[^\d]/) {                 # start or continuation of message?
            $msg .= $_;
        } elsif (m@^\d{4}/(?:0[1-9]|1[012])/($re_day) ($re_hour):($re_min)$@) {
            # time of report dissimination is N/A, this is the observation time
            $wmo_header = format_wmo_header "$wmo_header_fix$1$2$3", $opts->{H};
        }
    }
    print_metaf_msg $wmo_header, $msg, $opts
        if $msg;
    return;
}

use Getopt::Std;

sub usage {
    print STDERR
          (shift)
        . "Usage: metafsrc2raw.pl -F <format> [OPTION]... [FILE] ...\n"
        . "  -F <format>  format of the input data:\n"
        . "               metar_cycle_iws, staf_cycle_iws, taf_cycle_iws,\n"
        . "               metaf_nws, metaf_cod, metaf_fsu,\n"
        . "               synop_nws, synop_cod, synop_fsu\n"
        . "  Options:\n"
        . "  -D           suppress duplicate messages\n"
        . "  -N           suppress NIL messages\n"
        . "  -H <0..4>    which parts of the WMO header to include\n";
    exit 1;
}

sub main {
    my %opts;

    # get and check options
    usage '' unless getopts('F:DNH:', \%opts);
    usage "ERROR: input format must be specified with -F<FORMAT>.\n"
        unless exists $opts{F};
    $opts{H} = 0 unless exists $opts{H} && $opts{H} =~ /^[0-4]$/;

    if ($opts{F} =~ /^(metar|staf|taf)_cycle_iws$/) {
        process_metaf_cycle_iws(\%opts, $1);
    } elsif ($opts{F} =~ /^metaf_nws$/) {
        process_metaf_nws_cod_fsu(\%opts, '(.*)####\d{9}####$');
    } elsif ($opts{F} eq 'metaf_cod' || $opts{F} eq 'metaf_fsu') {
        process_metaf_nws_cod_fsu(\%opts, '^\d{3} ()$');
    } elsif ($opts{F} eq 'synop_nws') {
        process_synop_nws_cod_fsu(\%opts, '(.*)####\d{9}####$');
    } elsif ($opts{F} eq 'synop_cod' || $opts{F} eq 'synop_fsu') {
        process_synop_nws_cod_fsu(\%opts, '^\d{3} ()$');
    } else {
        usage "ERROR: invalid input data format: '" . $opts{F} . "'.\n";
    }
}

&main;
